import logging
import os
from typing import List
import proto_schema_parser.parser as pb_parser
from proto_schema_parser.ast import Package, Service, Message, Method, Field

from appsec_discovery.parsers import Parser
from appsec_discovery.models import CodeObject, CodeObjectField

logger = logging.getLogger(__name__)

class ProtobufParser(Parser):

    def find_proto_files(self, root_dir):
        proto_files = []
        for root, _, files in os.walk(root_dir):
            for file in files:
                if file.endswith('.proto'):
                    proto_files.append(os.path.join(root, file))
        return proto_files


    def run_scan(self) -> List[CodeObject]:

        objects_list: List[CodeObject] = []

        proto_files = self.find_proto_files(self.source_folder)
        proto_data = {}

        for proto_file in proto_files:

            local_proto_file = proto_file.replace(self.source_folder, "")

            try:
                with open(proto_file) as file:
                    file_str = file.read()
                    proto_data[local_proto_file] = pb_parser.Parser().parse(file_str)
            except Exception as ex:
                logger.error(f"Failed to parse {self.parser} for file {local_proto_file}: {ex}")

        objects_list = self.parse_report(proto_data)

        return objects_list

    def resolve_fields(self, type_name, messages, file):
        
        resolved_fields = {}

        local_messages = {}

        for el in messages[type_name].elements:       
            if isinstance(el, Message) and el.name not in local_messages:
                local_messages[el.name] = el

        for el in messages[type_name].elements:      
            if isinstance(el, Field) and el.name not in resolved_fields:

                if el.type in messages:
                    resolved_local_fields = self.resolve_fields(el.type, messages, file)

                    for field_name, field in resolved_local_fields.items():
                        resolved_fields[f"{type_name}.{el.name}.{field_name}"] = field

                elif el.type in local_messages:
                    resolved_local_fields = self.resolve_fields(el.type, local_messages, file)

                    for field_name, field in resolved_local_fields.items():
                        resolved_fields[f"{type_name}.{el.name}.{field_name}"] = field

                else:
                    resolved_fields[f"{type_name}.{el.name}"] = CodeObjectField(
                        field_name=f"{type_name}.{el.name}",
                        field_type=el.type,
                        file=file,
                        line=el.number
                    )

        if not resolved_fields :
            resolved_fields[type_name] = CodeObjectField(
                        field_name=type_name,
                        field_type='Empty',
                        file=file,
                        line=1
                    )

        return resolved_fields

    def parse_report(self, proto_data) -> List[CodeObject]:

        parsed_objects: List[CodeObject] = []

        packages = {}

        for file, file_proto in proto_data.items():

            cur_package = ''

            for el in file_proto.file_elements:

                if isinstance(el, Package) :

                    cur_package = el.name

                    if el.name not in packages:
                        packages[el.name] = {
                            'services': {},
                            'messages': {},
                            'file': file
                        }

            for el in file_proto.file_elements:

                if isinstance(el, Service) and el.name not in packages[cur_package]['services']:
                    packages[cur_package]['services'][el.name] = {}

                    for service_method in el.elements:
                        if isinstance(service_method, Method) and service_method.name not in packages[cur_package]['services'][el.name]:
                            packages[cur_package]['services'][el.name][service_method.name] = {
                                'input': service_method.input_type.type,
                                'output': service_method.output_type.type
                            }
                
                if isinstance(el, Message) and el.name not in packages[cur_package]['messages']:
                    packages[cur_package]['messages'][el.name] = el
        

        for package_name, package in packages.items():
            for service_name, service in package['services'].items():
                for method_name, method in service.items():

                    unique_hash = self.calc_uniq_hash([package['file'], package_name, service_name, method_name])

                    code_object = CodeObject(
                        hash=unique_hash,
                        object_name=f"Rpc /{package_name}.{service_name}/{method_name}",
                        object_type='rpc',
                        parser=self.parser,
                        file=package['file'],
                        line=1,
                        properties={},
                        fields={}
                    )

                    if method['input'] in package['messages']:
                        input_fields = self.resolve_fields(method['input'], package['messages'], package['file'])

                        for field_name, field in input_fields.items():    
                            code_object.fields[f"input.{field_name}"] = field
                    else:
                        code_object.fields['input'] = CodeObjectField(
                            field_name='input',
                            field_type=method['input'],
                            file=package['file'],
                            line=1
                        )

                    if method['output'] in package['messages']:
                        output_fields = self.resolve_fields(method['output'], package['messages'], package['file'])

                        for field_name, field in output_fields.items():    
                            code_object.fields[f"output.{field_name}"] = field
                    else:
                        code_object.fields['output'] = CodeObjectField(
                            field_name='output',
                            field_type=method['output'],
                            file=package['file'],
                            line=1
                        )

                    parsed_objects.append(code_object)

            '''
            for type_def in file_proto.definitions:

                if type_def.kind == 'object_type_extension':
                    extend_types[type_def.name.value] = 1

                if type_def.name.value not in types:

                    types[type_def.name.value] = {
                        'name': type_def.name.value,
                        'type': type_def.kind,
                        'file': file,
                        'line': type_def.loc.start,
                        'fields': {}
                    }

                if hasattr(type_def, 'fields'):
                    for field_def in type_def.fields:

                        output = ''

                        if hasattr(field_def.type, 'name'):
                            output = field_def.type.name.value
                        elif hasattr(field_def.type, 'type') and hasattr(field_def.type.type, 'name'):
                            output = field_def.type.type.name.value
                        elif hasattr(field_def.type, 'type') and hasattr(field_def.type.type, 'type') and hasattr(field_def.type.type.type, 'name'):
                            output = field_def.type.type.type.name.value
                        elif hasattr(field_def.type, 'type') and hasattr(field_def.type.type, 'type') and hasattr(field_def.type.type.type, 'type') and hasattr(field_def.type.type.type.type, 'name'):
                            output = field_def.type.type.type.type.name.value

                        if field_def.name.value not in types[type_def.name.value]['fields']:
                            types[type_def.name.value]['fields'][field_def.name.value] = {
                                'name': field_def.name.value,
                                'file': file,
                                'line': field_def.loc.start,
                                'inputs': {},
                                'output': output, 
                            }

                        if hasattr(field_def, 'arguments'):
                            for input_def in field_def.arguments:
                                if input_def.name.value not in types[type_def.name.value]['fields'][field_def.name.value]['inputs']:

                                    input_type = ''
                                    if hasattr(input_def.type, 'name'):
                                        input_type = input_def.type.name.value
                                    elif hasattr(input_def.type, 'type') and hasattr(input_def.type.type, 'name'):
                                        input_type = input_def.type.type.name.value
                                    elif hasattr(input_def.type, 'type') and hasattr(input_def.type.type, 'type') and hasattr(input_def.type.type.type, 'name'):
                                        input_type = input_def.type.type.type.name.value
                                    elif hasattr(input_def.type, 'type') and hasattr(input_def.type.type, 'type') and hasattr(input_def.type.type.type, 'type') and hasattr(input_def.type.type.type.type, 'name'):
                                        input_type = input_def.type.type.type.type.name.value

                                    types[type_def.name.value]['fields'][field_def.name.value]['inputs'][input_def.name.value] = input_type
    

        for type_name, type_dict in types.items():

            if type_name in extend_types :

                for field_name, field in type_dict['fields'].items():

                    if field['output'] not in extend_types:
            
                        proto_type = 'Query'
                        proto_name = f"{type_name}.{field_name}"

                        if type_name in ['Query', 'Mutation']:
                            proto_type = type_name
                            proto_name = field_name
                        
                        elif 'Mutation' in types and types['Mutation']['fields']:
                            for mutation_field in types['Mutation']['fields'].values():
                                if mutation_field['output'] == type_name:
                                    proto_type = 'Mutation'

                        result_fields = {}

                        for input_name, input_type in field['inputs'].items():

                            if input_type in types and types[input_type]['fields']:

                                inputs_resolved = self.resolve_fields(input_type, types)

                                for input_resolved_name, input_resolved in inputs_resolved.items():

                                    result_fields[f"{input_name}.{input_resolved_name}"] = input_resolved
                            else:

                                result_fields[f"{input_name}"] = {
                                    'type': input_type,
                                    'file': field['file'],
                                    'line': field['line'],
                                }

                        if field['output'] in types and types[field['output']]['fields'] :

                            outputs_resolved = self.resolve_fields(field['output'], types)

                            for output_resolved_name, output_resolved in outputs_resolved.items():

                                result_fields[f"output.{output_resolved_name}"] = output_resolved

                        else:
                            result_fields["output"] = {
                                'type': field['output'],
                                'file': field['file'],
                                'line': field['line'],
                            }

                        unique_hash = self.calc_uniq_hash([field_name, proto_type, field['file']])

                        code_object = CodeObject(
                            hash=unique_hash,
                            object_name=f"{proto_type} {proto_name}",
                            object_type=proto_type.lower(),
                            parser=self.parser,
                            file=field['file'],
                            line=field['line'],
                            properties={},
                            fields={}
                        )

                        for result_field_name, result_field in result_fields.items():
                            code_object.fields[result_field_name] = CodeObjectField(
                                field_name=result_field_name,
                                field_type=result_field['type'],
                                file=result_field['file'],
                                line=result_field['line']
                            )

                        parsed_objects.append(code_object)
        '''

        logger.info(f"For scan {self.parser} data parse {len(parsed_objects)} objects")

        return parsed_objects


from llama_cpp import Llama
import json
import yaml
import re

from typing import List, Dict
import logging

from appsec_discovery.models import CodeObject, ExcludeScoring, AiParams

severities_int = {'critical': 5, 'high': 4, 'medium': 3, 'low': 2, 'info': 1}
skip_ai = ['created_at', 'updated_at', 'deleted_at']

logger = logging.getLogger(__name__)

class AiService:

    def __init__(self, ai_params: AiParams, exclude_scoring: List[ExcludeScoring]):

        self.model_id = ai_params.model_id
        self.model_folder = ai_params.model_folder
        self.system_prompt = ai_params.prompt
        self.gguf_file = ai_params.gguf_file
        self.exclude_scoring = exclude_scoring
        

    def ai_score_objects(self, code_objects: List[CodeObject]) -> List[CodeObject]:

        scored_objects: List[CodeObject] = []

        try:

            llm = Llama.from_pretrained(
                repo_id=self.model_id,
                filename=self.gguf_file,
                verbose=False,
                cache_dir=self.model_folder
            )

            for object in code_objects:

                object_to_score: Dict[str: List[str]] = {'field_names': []}
                
                for field in object.fields.values() :

                    # no ai for already sensitive
                    if not field.severity \
                      and not ( field.field_name.split('.')[-1] in skip_ai \
                            or field.field_name.lower().endswith('id') \
                            or 'status' not in field.field_name.split('.')[-1].lower() ):
                        object_to_score['field_names'].append(field.field_name.split('.')[-1])
                
                object_text = f"object: {object.object_name}\n\n" + yaml.safe_dump(object_to_score)

                scored_list = []

                if object_to_score['field_names'] :
                
                    response = llm.create_chat_completion(
                        messages = [
                            {"role": "system", "content": self.system_prompt},
                            {"role": "user", "content": object_text },
                        ]
                    )

                    answer = response['choices'][0]["message"]["content"]

                    if '[' in answer and ']' in answer:

                        json_txt = "[" + answer.split('[')[1].split(']')[0] + "]"
                        scored_list = json.loads(json_txt)

                scored_fields = {}

                tag = "llm"
                severity = "medium"

                for field_name, field in object.fields.items():

                    if field.field_name.split('.')[-1] in scored_list:

                        excluded = False

                        for exclude in self.exclude_scoring:

                            if ( exclude.file or exclude.parser or exclude.object_name or exclude.object_type or exclude.prop_name 
                                    or exclude.field_name or exclude.field_type or exclude.tag or exclude.keyword ) \
                                and (exclude.parser is None or exclude.parser.lower() == object.parser.lower()) \
                                and (exclude.file is None or re.match(exclude.file, object.file) or exclude.file.lower() in object.file.lower()) \
                                and (exclude.object_name is None or re.match(exclude.object_name, object.object_name) or exclude.object_name.lower() in object.object_name.lower()) \
                                and (exclude.object_type is None or re.match(exclude.object_type, object.object_type) or exclude.object_type.lower() in object.object_type.lower()) \
                                and (exclude.prop_name is None ) \
                                and (exclude.field_name is None or re.match(exclude.field_name, field.field_name) or exclude.field_name.lower() in field.field_name.lower()) \
                                and (exclude.field_type is None or re.match(exclude.field_type, field.field_type) or exclude.field_type.lower() in field.field_type.lower()) \
                                and (exclude.tag is None or exclude.tag == tag ) \
                                and (exclude.keyword is None ) :
                                
                                excluded = True

                        if not excluded :

                            if not field.severity:
                                field.severity = severity
                                field.tags = [tag]
                            else:
                                if tag not in field.tags:
                                    field.tags.append(tag)

                                if severities_int[severity] > severities_int[field.severity]:
                                    field.severity = severity

                            if not object.severity:
                                object.severity = severity
                                object.tags = [tag]
                            else:
                                if tag not in object.tags:
                                    object.tags.append(tag)

                                if severities_int[severity] > severities_int[object.severity]:
                                    object.severity = severity

                    scored_fields[field_name] = field
                
                object.fields = scored_fields

                scored_objects.append(object)

        except Exception as ex:
            logger.error(f"Error while ai scoring: {ex}")

        # Check that all processed
        if len(scored_objects) == len(code_objects):
            return scored_objects
        
        else:
            return code_objects



import logging
import os
from typing import List, Dict
from pathlib import Path

from appsec_discovery.parsers import Parser
from appsec_discovery.models import CodeObject, CodeObjectProp

logger = logging.getLogger(__name__)

class TerraformParser(Parser):

    def run_scan(self) -> List[CodeObject]:

        objects_list: List[CodeObject] = []

        parser_folder = str(Path(__file__).resolve().parent)

        rules_folder = os.path.join(parser_folder, "scanner_rules")

        if os.path.isdir(rules_folder):
            
            semgrep_data = self.run_semgrep(source_folder=self.source_folder, rules_folder=rules_folder)
            objects_list = self.parse_report(semgrep_data)
        
        return objects_list


    def parse_report(self, semgrep_data) -> List[CodeObject]:

        parsed_objects: List[CodeObject] = []
        parsed_objects_dict: Dict[str, CodeObject] = {}

        for finding in semgrep_data:

            finding_file = finding.get('path')
            finding_line_int = finding.get('start',{}).get('line',0)

            rule_id = finding.get('check_id',"").split('.')[-1]

            vm_name = finding.get('extra').get('metavars').get('$...VM_NAME', {}).get('abstract_content',"")
            vm_domain = finding.get('extra').get('metavars').get('$...VM_DOMAIN', {}).get('abstract_content',"")
            vm_template = finding.get('extra').get('metavars').get('$...VM_TEMPLATE', {}).get('abstract_content',"")
            vm_pool = finding.get('extra').get('metavars').get('$...VM_POOL', {}).get('abstract_content',"")
            vm_desc = finding.get('extra').get('metavars').get('$...VM_DESC', {}).get('abstract_content',"")
            vm_server_cluster_name = finding.get('extra').get('metavars').get('$...VM_SERVER_CLUSTER_NAME', {}).get('abstract_content',"")
            vm_server_role = finding.get('extra').get('metavars').get('$...VM_SERVER_ROLE', {}).get('abstract_content',"")
            vm_server_owning_team = finding.get('extra').get('metavars').get('$...VM_SERVER_OWNING_TEAM', {}).get('abstract_content',"")
            vm_server_maintaining_team = finding.get('extra').get('metavars').get('$...VM_SERVER_MAINTAINING_TEAM', {}).get('abstract_content',"")
            vm_prometheus_env = finding.get('extra').get('metavars').get('$...VM_PROMETHEUS_ENV', {}).get('abstract_content',"")
            vlan_id = finding.get('extra').get('metavars').get('$...VLAN_ID', {}).get('abstract_content',"")
            dc = finding.get('extra').get('metavars').get('$...DC', {}).get('abstract_content',"")

            hash_key = self.calc_uniq_hash([finding_file, vm_name])

            if hash_key not in parsed_objects_dict:

                parsed_objects_dict[hash_key] = CodeObject(
                    hash=hash_key,
                    object_name=f"Virtual machine {vm_name}",
                    object_type="vm",
                    parser=self.parser,
                    file=finding_file,
                    line=finding_line_int,
                    properties={},
                    fields={}
                )

            if 'get-vm-name' in rule_id:
                parsed_objects_dict[hash_key].line = finding_line_int

            if vm_name :
                parsed_objects_dict[hash_key].properties['vm_name'] = CodeObjectProp(
                    prop_name='vm_name',
                    prop_value=vm_name
                )

            if vm_domain :
                parsed_objects_dict[hash_key].properties['vm_domain'] = CodeObjectProp(
                    prop_name='vm_domain',
                    prop_value=vm_domain
                )

                parsed_objects_dict[hash_key].object_name = f"Virtual machine {vm_name}.{vm_domain}"

            if vm_template :
                parsed_objects_dict[hash_key].properties['vm_template'] = CodeObjectProp(
                    prop_name='vm_template',
                    prop_value=vm_template
                ) 

            if vm_pool :
                parsed_objects_dict[hash_key].properties['vm_pool'] = CodeObjectProp(
                    prop_name='vm_pool',
                    prop_value=vm_pool
                )

            if vm_desc :
                parsed_objects_dict[hash_key].properties['vm_desc'] = CodeObjectProp(
                    prop_name='vm_desc',
                    prop_value=vm_desc
                )

            if vm_server_cluster_name :
                parsed_objects_dict[hash_key].properties['vm_server_cluster_name'] = CodeObjectProp(
                    prop_name='vm_server_cluster_name',
                    prop_value=vm_server_cluster_name
                )
                
            if vm_server_role :
                parsed_objects_dict[hash_key].properties['vm_server_role'] = CodeObjectProp(
                    prop_name='vm_server_role',
                    prop_value=vm_server_role
                )

            if vm_server_owning_team :
                parsed_objects_dict[hash_key].properties['vm_server_owning_team'] = CodeObjectProp(
                    prop_name='vm_server_owning_team',
                    prop_value=vm_server_owning_team
                )

            if vm_server_maintaining_team :
                parsed_objects_dict[hash_key].properties['vm_server_maintaining_team'] = CodeObjectProp(
                    prop_name='vm_server_maintaining_team',
                    prop_value=vm_server_maintaining_team
                )

            if vm_prometheus_env :
                parsed_objects_dict[hash_key].properties['vm_prometheus_env'] = CodeObjectProp(
                    prop_name='vm_prometheus_env',
                    prop_value=vm_prometheus_env
                )

            if vlan_id :
                parsed_objects_dict[hash_key].properties['vlan_id'] = CodeObjectProp(
                    prop_name='vlan_id',
                    prop_value=vlan_id
                )

            if dc :
                parsed_objects_dict[hash_key].properties['dc'] = CodeObjectProp(
                    prop_name='dc',
                    prop_value=dc
                )
        
        if parsed_objects_dict:
            parsed_objects = list(parsed_objects_dict.values())

        logger.info(f"For scan {self.parser} data parse {len(parsed_objects)} objects")
        return parsed_objects


import click

from appsec_discovery.services import ScanService, ReportService

@click.command()
@click.option('--source', required=True, type=click.Path(exists=True), help='Source code folder')
@click.option('--config', required=False, show_default=True, default=None, type=click.File('r'), help='Scoring config file')
@click.option('--output', required=False, show_default=True, default=None, type=click.File('w'), help='Output file')
@click.option('--output-type', required=False, show_default=True, default='yaml', type=click.Choice(['json', 'sarif', 'yaml'], case_sensitive=False), help='Report type')
@click.option("--only-scored-objects", is_flag=True, show_default=True, default=False, help="Show only scored objects")
def main(source, config, output, output_type, only_scored_objects):

    scan_service = ScanService(source_folder=source, conf_file=config, only_scored_objects=only_scored_objects)
    scanned_objects = scan_service.scan_folder()

    report_service = ReportService(code_objects=scanned_objects, report_type=output_type, report_file=output)
    report_service.save_report_to_disk()

if __name__ == '__main__':
    main()



import logging
import os
from typing import List
import graphql

from appsec_discovery.parsers import Parser
from appsec_discovery.models import CodeObject, CodeObjectField

logger = logging.getLogger(__name__)

class GraphqlParser(Parser):

    def find_graphql_files(self, root_dir):
        proto_files = []
        for root, _, files in os.walk(root_dir):
            for file in files:
                if file.endswith('.graphql'):
                    proto_files.append(os.path.join(root, file))
        return proto_files


    def run_scan(self) -> List[CodeObject]:

        objects_list: List[CodeObject] = []

        gql_files = self.find_graphql_files(self.source_folder)
        gql_data = {}

        for gql_file in gql_files:

            local_gql_file = gql_file.replace(self.source_folder, "")

            try:
                with open(gql_file) as file:
                    file_str = file.read()
                    gql_data[local_gql_file] = graphql.parse(file_str, no_location=False)
            except Exception as ex:
                logger.error(f"Failed to parse {self.parser} for file {local_gql_file}: {ex}")

        objects_list = self.parse_report(gql_data)

        return objects_list

    def resolve_fields(self, type_name, types_dict):
        
        resolved_fields = {}

        for field_name, field in types_dict[type_name]['fields'].items():

            if field['output'] in types_dict and types_dict[field['output']]['fields'] :

                out_resolved = self.resolve_fields(field['output'], types_dict)

                for out_field_name, out_field in out_resolved.items():
                    resolved_fields[f"{type_name}.{field_name}.{field['output']}.{out_field_name}"] = out_field

            else:
                resolved_fields[f"{type_name}.{field_name}"] = {
                    'type': field['output'],
                    'file': field['file'],
                    'line': field['line'],
                } 

        return resolved_fields

    def parse_report(self, gql_data) -> List[CodeObject]:

        parsed_objects: List[CodeObject] = []

        extend_types = {}

        types = {}

        for file, file_gql in gql_data.items():

            for type_def in file_gql.definitions:

                if type_def.kind == 'object_type_extension':
                    extend_types[type_def.name.value] = 1

                if type_def.name.value not in types:

                    types[type_def.name.value] = {
                        'name': type_def.name.value,
                        'type': type_def.kind,
                        'file': file,
                        'line': type_def.loc.start,
                        'fields': {}
                    }

                if hasattr(type_def, 'fields'):
                    for field_def in type_def.fields:

                        output = ''

                        if hasattr(field_def.type, 'name'):
                            output = field_def.type.name.value
                        elif hasattr(field_def.type, 'type') and hasattr(field_def.type.type, 'name'):
                            output = field_def.type.type.name.value
                        elif hasattr(field_def.type, 'type') and hasattr(field_def.type.type, 'type') and hasattr(field_def.type.type.type, 'name'):
                            output = field_def.type.type.type.name.value
                        elif hasattr(field_def.type, 'type') and hasattr(field_def.type.type, 'type') and hasattr(field_def.type.type.type, 'type') and hasattr(field_def.type.type.type.type, 'name'):
                            output = field_def.type.type.type.type.name.value

                        if field_def.name.value not in types[type_def.name.value]['fields']:
                            types[type_def.name.value]['fields'][field_def.name.value] = {
                                'name': field_def.name.value,
                                'file': file,
                                'line': field_def.loc.start,
                                'inputs': {},
                                'output': output, 
                            }

                        if hasattr(field_def, 'arguments'):
                            for input_def in field_def.arguments:
                                if input_def.name.value not in types[type_def.name.value]['fields'][field_def.name.value]['inputs']:

                                    input_type = ''
                                    if hasattr(input_def.type, 'name'):
                                        input_type = input_def.type.name.value
                                    elif hasattr(input_def.type, 'type') and hasattr(input_def.type.type, 'name'):
                                        input_type = input_def.type.type.name.value
                                    elif hasattr(input_def.type, 'type') and hasattr(input_def.type.type, 'type') and hasattr(input_def.type.type.type, 'name'):
                                        input_type = input_def.type.type.type.name.value
                                    elif hasattr(input_def.type, 'type') and hasattr(input_def.type.type, 'type') and hasattr(input_def.type.type.type, 'type') and hasattr(input_def.type.type.type.type, 'name'):
                                        input_type = input_def.type.type.type.type.name.value

                                    types[type_def.name.value]['fields'][field_def.name.value]['inputs'][input_def.name.value] = input_type
    

        for type_name, type_dict in types.items():

            if type_name in extend_types :

                for field_name, field in type_dict['fields'].items():

                    if field['output'] not in extend_types:
            
                        gql_type = 'Query'
                        gql_name = f"{type_name}.{field_name}"

                        if type_name in ['Query', 'Mutation']:
                            gql_type = type_name
                            gql_name = field_name
                        
                        elif 'Mutation' in types and types['Mutation']['fields']:
                            for mutation_field in types['Mutation']['fields'].values():
                                if mutation_field['output'] == type_name:
                                    gql_type = 'Mutation'

                        result_fields = {}

                        for input_name, input_type in field['inputs'].items():

                            if input_type in types and types[input_type]['fields']:

                                inputs_resolved = self.resolve_fields(input_type, types)

                                for input_resolved_name, input_resolved in inputs_resolved.items():

                                    result_fields[f"{input_name}.{input_resolved_name}"] = input_resolved
                            else:

                                result_fields[f"{input_name}"] = {
                                    'type': input_type,
                                    'file': field['file'],
                                    'line': field['line'],
                                }

                        if field['output'] in types and types[field['output']]['fields'] :

                            outputs_resolved = self.resolve_fields(field['output'], types)

                            for output_resolved_name, output_resolved in outputs_resolved.items():

                                result_fields[f"output.{output_resolved_name}"] = output_resolved

                        else:
                            result_fields["output"] = {
                                'type': field['output'],
                                'file': field['file'],
                                'line': field['line'],
                            }

                        unique_hash = self.calc_uniq_hash([field_name, gql_type, field['file']])

                        code_object = CodeObject(
                            hash=unique_hash,
                            object_name=f"{gql_type} {gql_name}",
                            object_type=gql_type.lower(),
                            parser=self.parser,
                            file=field['file'],
                            line=field['line'],
                            properties={},
                            fields={}
                        )

                        for result_field_name, result_field in result_fields.items():
                            code_object.fields[result_field_name] = CodeObjectField(
                                field_name=result_field_name,
                                field_type=result_field['type'],
                                file=result_field['file'],
                                line=result_field['line']
                            )

                        parsed_objects.append(code_object)

        logger.info(f"For scan {self.parser} data parse {len(parsed_objects)} objects")

        return parsed_objects


from pydantic import BaseModel
from typing import List, Dict, Optional


class CodeObjectField(BaseModel):
    field_name: str
    field_type: str
    file: Optional[str]
    line: Optional[int]
    severity: Optional[str]
    tags: Optional[List[str]]

class CodeObjectProp(BaseModel):
    prop_name: str
    prop_value: str
    file: Optional[str]
    line: Optional[int]
    severity: Optional[str]
    tags: Optional[List[str]]

class CodeObject(BaseModel):
    hash: str
    object_name: str
    object_type: str
    parser: str
    severity: Optional[str]
    tags: Optional[List[str]]
    file: str
    line: int
    properties: Dict[str, CodeObjectProp]
    fields: Dict[str, CodeObjectField]





from appsec_discovery.parsers.base_parser import Parser

import logging
import os
from importlib import import_module
from importlib.util import find_spec
from inspect import isclass
from pathlib import Path
from typing import Optional, List

logger = logging.getLogger(__name__)

class ParserFactory:

    @staticmethod
    def get_parser(parser_type: str):

        package_dir = str(Path(__file__).resolve().parent)

        for module_name in os.listdir(package_dir):

            if os.path.isdir(os.path.join(package_dir, module_name)) and parser_type == module_name:

                try:
                    # check if it's a Python module
                    if find_spec(f"appsec_discovery.parsers.{module_name}.parser"):

                        module = import_module(f"appsec_discovery.parsers.{module_name}.parser")

                        for attribute_name in dir(module):

                            attribute = getattr(module, attribute_name)
                            if isclass(attribute) and 'parser' in attribute_name.lower() and attribute_name != 'Parser':

                                parser = attribute
                                logger.info(f"Found parser {parser.__qualname__} in appsec_discovery.parsers.{module_name}.parser")
                                
                                return parser
                except Exception as ex:
                    logger.error(f"Failed to find parser for type {parser_type}: {ex}")

        logger.error(f"Parser for type {parser_type} not found")
        return None
    
    @staticmethod
    def get_parser_types() -> List[str]:

        parsers_list = []

        package_dir = str(Path(__file__).resolve().parent)

        for module_name in os.listdir(package_dir):

            if os.path.isdir(os.path.join(package_dir, module_name)) :

                try:
                    # check if it's a Python module
                    if find_spec(f"appsec_discovery.parsers.{module_name}.parser"):

                        module = import_module(f"appsec_discovery.parsers.{module_name}.parser")

                        for attribute_name in dir(module):

                            attribute = getattr(module, attribute_name)
                            if isclass(attribute) and 'parser' in attribute_name.lower() and attribute_name != 'Parser':
                                parsers_list.append(module_name)

                except:
                    logger.exception(f"failed to load {module_name}")

        return parsers_list


from pydantic import BaseModel
from typing import List, Dict, Optional

default_rules = {
    'pii': {
       'high': [
          'full_name',
          'fullname',
          'first_name',
          'firstname',
          'last_name',
          'lastname',
          'phone',
          'passport',
          'ssn'
       ],
       'medium': [
          'address',
          'email'
       ],
       'low': [
          'city'
       ]
    },
    'finance': {
       'high': [
          'pan',
          'card_number'
          'cardnumber'
       ],
       'medium': [
          'account_number',
          'accountnumber',
          'balance',
          'amount'
       ],
       'low': [
       ]
    },
    'auth': {
       'high': [
          'password',
          'pincode',
          'codeword',
          'token'
       ],
       'medium': [
          'login',
          'email',
          'phone',
          'username',
       ],
       'low': [
       ]
    },
}

class ExcludeScan(BaseModel):
    file: Optional[str]
    parser: Optional[str]
    object_name: Optional[str]
    object_type: Optional[str]

class ExcludeScoring(BaseModel):
    file: Optional[str]
    parser: Optional[str]
    object_name: Optional[str]
    object_type: Optional[str]
    prop_name: Optional[str]
    field_name: Optional[str]
    field_type: Optional[str]
    tag: Optional[str]
    keyword: Optional[str]

class AiParams(BaseModel):
    model_id: str
    model_folder: str
    gguf_file: str
    prompt: str

class ScoreConfig(BaseModel):
    parsers: List[str] = ['all']
    object_types: List[str] = ['all']
    score_tags: Dict[str,Dict[str,List[str]]] = default_rules
    ai_params: Optional[AiParams]
    exclude_scan: List[ExcludeScan] = []
    exclude_scoring: List[ExcludeScoring] = []




import logging
import os
import copy
from typing import List, Dict
from pathlib import Path

from appsec_discovery.parsers import Parser
from appsec_discovery.models import CodeObject, CodeObjectProp, CodeObjectField

logger = logging.getLogger(__name__)

class SwaggerParser(Parser):

    def run_scan(self) -> List[CodeObject]:

        objects_list: List[CodeObject] = []

        parser_folder = str(Path(__file__).resolve().parent)

        rules_folder = os.path.join(parser_folder, "scanner_rules")

        if os.path.isdir(rules_folder):
            
            semgrep_data = self.run_semgrep(source_folder=self.source_folder, rules_folder=rules_folder)
            objects_list = self.parse_report(semgrep_data)
        
        return objects_list


    def parse_report(self, semgrep_data) -> List[CodeObject]:

        parsed_objects: List[CodeObject] = []
        parsed_objects_dict: Dict[str, CodeObject] = {}
        parsed_fields_dict = {}

        objects_dict = {}
        routes_dict = {}

        for finding in semgrep_data:

            finding_file = finding.get('path')
            finding_line_int = finding.get('start',{}).get('line',0)

            rule_id = finding.get('check_id',"").split('.')[-1]

            if rule_id == 'swagger-schema':

                object = finding.get('extra').get('metavars').get('$OBJECT', {}).get('abstract_content',"")
                field_name = finding.get('extra').get('metavars').get('$FIELD', {}).get('abstract_content',"")
                field_type = finding.get('extra').get('metavars').get('$TYPE', {}).get('abstract_content',"")
                field_ref = finding.get('extra').get('metavars').get('$REF', {}).get('abstract_content',"").split('/')[-1]

                if object not in objects_dict:
                    objects_dict[object] = {}
                
                if field_name not in objects_dict[object]:
                    objects_dict[object][field_name] = {'field_type': field_type, 'ref': field_ref}

                if field_ref and objects_dict[object][field_name]['field_type'] != 'ref':
                    objects_dict[object][field_name]['field_type'] = 'ref'
                    objects_dict[object][field_name]['ref'] = field_ref

        # resolving refs

        no_changes = False
        while not no_changes:

            no_changes = True
            objects_dict_copy = copy.deepcopy(objects_dict)

            for obj, obj_fields in objects_dict_copy.items():
                for field, field_data in obj_fields.items():

                    if field_data['ref'] in objects_dict_copy:

                        for ref_field, ref_field_data in objects_dict_copy[field_data['ref']].items():
                            objects_dict[obj][f"{field}.{ref_field}"] = ref_field_data

                        del objects_dict[obj][field]
                        no_changes = False

        for finding in semgrep_data:

            finding_file = finding.get('path')
            finding_line_int = finding.get('start',{}).get('line',0)

            rule_id = finding.get('check_id',"").split('.')[-1]

            if rule_id != 'swagger-schema':

                path = finding.get('extra').get('metavars').get('$PATH', {}).get('abstract_content',"").strip('"').strip("'")
                method = finding.get('extra').get('metavars').get('$METHOD', {}).get('abstract_content',"").strip('"').strip("'")

                field_name = finding.get('extra').get('metavars').get('$PARAM', {}).get('abstract_content',"").strip('"').strip("'")
                field_type = finding.get('extra').get('metavars').get('$TYPE', {}).get('abstract_content',"").strip('"').strip("'")

                req_ref = finding.get('extra').get('metavars').get('$REQREF', {}).get('abstract_content',"").split('/')[-1]
                resp_ref = finding.get('extra').get('metavars').get('$RESPREF', {}).get('abstract_content',"").split('/')[-1]

                hash_key = self.calc_uniq_hash([finding_file, path, method])
                field_hash_key = self.calc_uniq_hash([finding_file, path, method, field_name])
                req_ref_hash_key = self.calc_uniq_hash([finding_file, path, method, req_ref, 'req'])
                resp_ref_hash_key = self.calc_uniq_hash([finding_file, path, method, resp_ref, 'resp'])

                object_name = f"Swagger route {path} ({method.upper()})"

                if hash_key not in parsed_objects_dict:

                    parsed_objects_dict[hash_key] = CodeObject(
                        hash=hash_key,
                        object_name=object_name,
                        object_type="route",
                        parser=self.parser,
                        file=finding_file,
                        line=finding_line_int,
                        properties={},
                        fields={}
                    )

                    parsed_objects_dict[hash_key].properties['path'] = CodeObjectProp(
                        prop_name='path',
                        prop_value=path
                    )

                    parsed_objects_dict[hash_key].properties['method'] = CodeObjectProp(
                        prop_name='method',
                        prop_value=method
                    )

                if field_name and field_hash_key not in parsed_fields_dict:

                    parsed_field = CodeObjectField(
                        field_name=f"Input.{field_name}",
                        field_type=field_type,
                        file=finding_file,
                        line=finding_line_int
                    )

                    parsed_objects_dict[hash_key].fields[parsed_field.field_name] = parsed_field
                    parsed_fields_dict[field_hash_key] = parsed_field

                if req_ref in objects_dict and req_ref_hash_key not in parsed_fields_dict:
                    for ref_field, ref_field_data in objects_dict[req_ref].items():

                        parsed_field = CodeObjectField(
                            field_name=f"Input.{req_ref}.{ref_field}",
                            field_type=ref_field_data['field_type'],
                            file=finding_file,
                            line=finding_line_int
                        )

                        parsed_objects_dict[hash_key].fields[parsed_field.field_name] = parsed_field
                        parsed_fields_dict[req_ref_hash_key] = parsed_field

                if resp_ref in objects_dict and resp_ref_hash_key not in parsed_fields_dict:
                    for ref_field, ref_field_data in objects_dict[resp_ref].items():

                        parsed_field = CodeObjectField(
                            field_name=f"Output.{resp_ref}.{ref_field}",
                            field_type=ref_field_data['field_type'],
                            file=finding_file,
                            line=finding_line_int
                        )

                        parsed_objects_dict[hash_key].fields[parsed_field.field_name] = parsed_field
                        parsed_fields_dict[resp_ref_hash_key] = parsed_field

        if parsed_objects_dict:
            parsed_objects = list(parsed_objects_dict.values())

        logger.info(f"For scan {self.parser} data parse {len(parsed_objects)} objects")
        return parsed_objects